/***********************************************************************
* FILENAME :        gpc.c 
*
* DESCRIPTION :
*        GPC algorithm implementation
*
*
* licensed under the Creative Commons Attribution-NonCommercial 4.0 
* 
* AUTHOR :    Romulo Antao        START DATE :    27 MAR 17
*
************************************************************************/

/* Standard includes. */
#include <stdio.h>

/* Hardware includes. */
#include <math.h> 

/* Misc definitions */
#include "gpc.h"
#include "t2fls.h"

/* GPC internal variables */
Matrix G;
Matrix L;
Matrix f;
Matrix g;

Matrix K;
Matrix y;
Matrix r;
Matrix u;

/* Auxiliary matrices*/
Matrix K_inv;
Matrix K_tmp;
Matrix Fy_tmp;

/* Initialize GPC Matrices */
int gpc_init(void)
{
	matrix_init(&G, Np, Nu);
	matrix_init(&L, Np, 1);
	matrix_init(&f, Np, na+1);
	matrix_init(&g, Np, ng);

	matrix_init(&K, 1, Np);

	matrix_init(&y, na+1, 1);
	matrix_init(&u, nb, 1);
	matrix_init(&r, Np, 1);


	matrix_init(&K_tmp, Nu, Nu);
	matrix_init(&K_inv, Nu, Nu);
	matrix_init(&Fy_tmp,Np,1);

	return 0;
}

/* Implementation of the linear GPC */
float gpc(Matrix *a_tilde, Matrix *b, float du, float penalty)
{

	int i=0, j=0, p=0, n=0;
	float new_du = 0;

	// Get the f coefficients (free response) -> f[Np,na]
	for(i=0;i<na+1;i++)
	{
		f.matrix[0][i] = - a_tilde->matrix[i+1][0];
	}

	for(p=1;p<Np;p++)
	{
		for(i=0;i<na;i++)
		{
			f.matrix[p][i] = f.matrix[p-1][i+1] - f.matrix[p-1][0]*a_tilde->matrix[i+1][0];
		}
		f.matrix[p][na] = - f.matrix[p-1][0]*a_tilde->matrix[na+1][0];
	}

	// Get the g coefficients (impulse response) -> g[Np,Np+nb-1]
	for(i=0;i<nb;i++)
	{
		g.matrix[0][i] = b->matrix[i][0]; 
	}

	for(p=1;p<Np;p++)
	{
		for(i=0;i<p;i++)
		{
			g.matrix[p][i] = g.matrix[p-1][i];
		}
		
		for(i=0;i<nb;i++)
		{
			g.matrix[p][p+i] = g.matrix[p-1][p+i] + f.matrix[p-1][0]*b->matrix[i][0]; 
		}
	}

	// Get the G Matrix
	for(p=0;p<Np;p++)
	{
		i=p;
		j=0;
		while((i>=0) && (j<Nu))
		{
			G.matrix[p][j] = g.matrix[p][i];
			j++;
			i--;
		}
	}

	// Get the L matrix
	for(p=0;p<Np;p++)
	{
		L.matrix[p][0] = g.matrix[p][p+1];
	}

	// Calculate the GPC gain matrix K
	matrix_MTM(&G, &K_inv);
	matrix_sumdiag(&K_inv, penalty, &K_inv);
 	//Check the size of Nu!! MUST be concordant with the dimension of the inverted matrix Nu=3 -> M3x3 
	matrix_inv3x3(&K_inv, &K_tmp);
	gpc_K();

	// Calculate F*y
	matrix_prod(&f, &y, &Fy_tmp);

	// Calculate L*du
	matrix_scale(&L, du, &L);

	// Calculate (r - F*y - L*du)
	matrix_minus(&r, &Fy_tmp, &Fy_tmp);
	matrix_minus(&Fy_tmp, &L, &Fy_tmp);

	// Calculate the incremental control signal ( du=K*(r - F*y - L*du) )
	new_du = gpc_du();

	return new_du;
}

/* calculates the first row of the GPC gain matrix K (K=K_tmp*G^T) */
int gpc_K(void)
{
	int i, j, k;

	for (j=0;j<G.rows;j++)
	{
		K.matrix[0][j] = 0.0;
		for (k=0;k<G.cols;k++)
		{
			K.matrix[0][j] += K_tmp.matrix[0][k]*G.matrix[j][k];
		}
	}

	K.rows = 1;
	K.cols = G.rows;
	return 0;
}

/* calculate the gpc output value */
float gpc_du(void)
{
	int i, j;
	float du = 0.0;

	for (i=0;i<K.cols;i++)
	{
		du += K.matrix[0][i]*Fy_tmp.matrix[i][0];
	}
	return du;
}

/* shifts new data to the first position of the y vector (column vector) */
int gpc_yShift(float new_data)
{
	int i;

	for (i=y.rows-1;i>0;i--)
	{
		y.matrix[i][0] = y.matrix[i-1][0];
	}
	y.matrix[0][0] = new_data;
	return 0;
}

/* shifts new data to the first position of the y vector (column vector) */
int gpc_uShift(float new_data)
{
	int i;

	for (i=u.rows-1;i>0;i--)
	{
		u.matrix[i][0] = u.matrix[i-1][0];
	}
	u.matrix[0][0] = new_data;
	return 0;
}

/* shifts new data to the last position of the reference vector (column vector) */
int gpc_rShift(float new_data)
{
	int i;

	for (i=0;i<r.rows-1;i++)
	{
		r.matrix[i][0] = r.matrix[i+1][0];
	}
	r.matrix[r.rows-1][0] = new_data;
	return 0;
}

/* initialize the regression variables */
int gpc_regInit(float r_init[], float y_init[], float u_init[])
{
	int i;

	for (i=0;i<Np;i++)
	{
		r.matrix[i][0] = r_init[i];
	}

	for (i=0;i<na+1;i++)
	{
		y.matrix[i][0] = y_init[i];
	}

	for (i=0;i<nb;i++)
	{
		u.matrix[i][0] = u_init[i];
	}

	return 0;
}