/***********************************************************************
* FILENAME :        matrix.c 
*
* DESCRIPTION :
*        Matrix operations
*
* NOTES:
* 	     Warning! malloc defined for freeRTOS
*
* licensed under the Creative Commons Attribution-NonCommercial 4.0 
* 
* AUTHOR :    Romulo Antao        START DATE :    27 MAR 17
*
************************************************************************/

/* Standard includes. */
#include <stdio.h>
#include <math.h>

/* Misc Includes */
#include "matrix.h"
#include "FreeRTOS.h"

/* Function prototypes */
int matrix_init(Matrix *Mdst, int rows, int columns);

int matrix_MTM(Matrix *Msrc, Matrix *Mdst);
int matrix_sumdiag(Matrix *Msrc, float var, Matrix *Mdst);
int matrix_inv3x3(Matrix *Msrc, Matrix *Mdst);
int matrix_inv2x2(Matrix *Msrc, Matrix *Mdst);
int matrix_minus(Matrix *Msrc, Matrix *M2src, Matrix *Mdst);
int matrix_sum(Matrix *M1src, Matrix *M2src, Matrix *Mdst);
int matrix_prod(Matrix *M1src, Matrix *M2src, Matrix *Mdst);
int matrix_scale(Matrix *Msrc, float var, Matrix *Mdst);

/* initialize the matrix structure and zero it*/
int matrix_init(Matrix *Mdst, int rows, int columns)
{
	int i, j;

	/* Reserve memory */
	Mdst->matrix = pvPortMalloc(rows*sizeof(float));

	for (i=0;i<rows;i++)
	{
		Mdst->matrix[i] = pvPortMalloc(columns*sizeof(float));
	}

	/* Set the matrix dimensions */
	Mdst->rows = rows;
	Mdst->cols = columns;
	
	/* Zeroing the matrix */
	for (i=0;i<Mdst->rows;i++)
	{
		for (j=0;j<Mdst->cols;j++)
		{
			Mdst->matrix[i][j] = 0.0;
		}
	}

	return 0;
}

/* calculates the square matrix resulting from M^T*M */
int matrix_MTM(Matrix *Msrc, Matrix *Mdst)
{
	int i, j, k;

	for (i=0;i<Msrc->cols;i++)
	{
		for (j=0;j<=i;j++)
		{
			Mdst->matrix[i][j] = 0.0;

			for (k=0;k<Msrc->rows;k++)
			{
				Mdst->matrix[i][j] += Msrc->matrix[k][i]*Msrc->matrix[k][j];
			}
			/* Since the matrix is symmetric */
			Mdst->matrix[j][i] = Mdst->matrix[i][j];

			Mdst->rows = Msrc->cols;
			Mdst->cols = Msrc->cols;
		}
	}
	return 0;
}

/* sum scalar to matrix diagonal */
int matrix_sumdiag(Matrix *Msrc, float var, Matrix *Mdst)
{
	int i;	

	// check if matrix is square
	if (Msrc->rows==Msrc->cols)
	{
		for (i=0;i<Msrc->rows;i++)
		{
				Mdst->matrix[i][i] = Msrc->matrix[i][i]+var;

				Mdst->rows = Msrc->rows;
				Mdst->cols = Msrc->cols;
		}
	}
	else
	{
		return -1; // matrix is not square
	}

	return 0;
}

/* calculate the inverse of a 3x3 symmetric matrix */
int matrix_inv3x3(Matrix *Msrc, Matrix *Mdst)
{	
	float aux_1;	 // Msrc[2][2] * Msrc[1][1] - Msrc[2][1] * Msrc[1][2]
	float aux_2;	 // Msrc[2][2] * Msrc[0][1] - Msrc[2][1] * Msrc[0][2]
	float aux_3;	 // Msrc[1][2] * Msrc[0][1] - Msrc[1][1] * Msrc[0][2]
	float det, idet; // Matrix determinant and its inverse

	// calculate useful products
	aux_1 = Msrc->matrix[1][1] * Msrc->matrix[2][2] - Msrc->matrix[1][2] * Msrc->matrix[2][1];
	aux_2 = Msrc->matrix[0][1] * Msrc->matrix[2][2] - Msrc->matrix[1][2] * Msrc->matrix[0][2];
	aux_3 = Msrc->matrix[0][1] * Msrc->matrix[1][2] - Msrc->matrix[1][1] * Msrc->matrix[0][2];

	// set det to the determinant of the input matrix Msrc
	det = Msrc->matrix[0][0] * aux_1 - Msrc->matrix[1][0] * aux_2 + Msrc->matrix[2][0] * aux_3;

	// set Mdst to the inverse of Msrc for any determinant except zero
	if (det != 0.0)
	{
		idet = 1.0 / det;
		Mdst->matrix[0][0] = aux_1 * idet;
		Mdst->matrix[1][0] = Mdst->matrix[0][1] = -aux_2 * idet;
		Mdst->matrix[2][0] = Mdst->matrix[0][2] = aux_3 * idet;
		Mdst->matrix[1][1] = (Msrc->matrix[0][0] * Msrc->matrix[2][2] - Msrc->matrix[0][2] * Msrc->matrix[2][0]) * idet;
		Mdst->matrix[2][1] = Mdst->matrix[1][2] = (Msrc->matrix[0][2] * Msrc->matrix[0][1] - Msrc->matrix[0][0] * Msrc->matrix[1][2]) * idet;
		Mdst->matrix[2][2] = (Msrc->matrix[0][0] * Msrc->matrix[1][1] - Msrc->matrix[0][1] * Msrc->matrix[1][0]) * idet;

		Mdst->rows = Msrc->rows;
		Mdst->cols = Msrc->cols;
	}
	else
	{
		return -1; //Matrix is not invertible (determinant is zero)
	}
	return 0;
}

/* calculate the inverse of a 2x2 symmetric matrix */
int matrix_inv2x2(Matrix *Msrc, Matrix *Mdst)
{	
	float det, idet;	// Matrix determinant and its inverse

	// set det to the determinant of the input matrix Msrc
	det = Msrc->matrix[0][0] * Msrc->matrix[1][1] - Msrc->matrix[0][1] * Msrc->matrix[1][0];

	// set Mdst to the inverse of Msrc for any determinant except zero
	if (det != 0.0)
	{
		idet = 1.0 / det;
		Mdst->matrix[0][0] = Msrc->matrix[1][1] * idet;
		Mdst->matrix[1][0] = Mdst->matrix[0][1] = -Msrc->matrix[0][1] * idet;
		Mdst->matrix[1][1] = Msrc->matrix[0][0] * idet;

		Mdst->rows = Msrc->rows;
		Mdst->cols = Msrc->cols;		
	}
	else
	{
		return -1; //Matrix is not invertible (determinant is zero)
	}
	return 0;
}

/* subtract two matrices */
int matrix_minus(Matrix *M1src, Matrix *M2src, Matrix *Mdst)
{
	int i, j;

	if ((M1src->rows == M2src->rows) && (M1src->cols == M2src->cols))
	{
		for (i=0;i<M1src->rows;i++)
		{
			for (j=0;j<M1src->cols;j++)
			{
				Mdst->matrix[i][j] = M1src->matrix[i][j] - M2src->matrix[i][j];
			}
		}
		Mdst->rows = M1src->rows;
		Mdst->cols = M1src->cols;	
	}
	else
	{
		return -1;
	}
	return 0;

}

/* sum two matrices */
int matrix_sum(Matrix *M1src, Matrix *M2src, Matrix *Mdst)
{
	int i, j;

	if ((M1src->rows == M2src->rows) && (M1src->cols == M2src->cols))
	{
		for (i=0;i<M1src->rows;i++)
		{
			for (j=0;j<M1src->cols;j++)
			{
				Mdst->matrix[i][j] = M1src->matrix[i][j] + M2src->matrix[i][j];
			}
		}
		Mdst->rows = M1src->rows;
		Mdst->cols = M1src->cols;	
	}
	else
	{
		return -1;
	}
	return 0;

}

/* multiply two matrices */
int matrix_prod(Matrix *M1src, Matrix *M2src, Matrix *Mdst)
{
	int i, j, k;

	// check if inner dimensions are equal
	if (M1src->cols==M2src->rows)
	{
		for (i=0;i<M1src->rows;i++)
		{
			for (j=0;j<M2src->cols;j++)
			{
				Mdst->matrix[i][j] = 0.0;
				for (k=0;k<M1src->cols;k++)
					Mdst->matrix[i][j] += M1src->matrix[i][k]*M2src->matrix[k][j];
			}
		}
		Mdst->rows = M1src->rows;
		Mdst->cols = M2src->cols;
	}
	else
	{
		return -1; // matrix product cannot be applied
	}
	return 0;
}

int matrix_scale(Matrix *Msrc, float var, Matrix *Mdst)
{
	int i, j;

	for (i=0;i<Msrc->rows;i++)
	{
		for (j=0;j<Msrc->cols;j++)
		{
			Mdst->matrix[i][j] = Msrc->matrix[i][j]*var;
		}
	}
	Mdst->rows = Msrc->rows;
	Mdst->cols = Msrc->cols;

	return 0;
}


