/***********************************************************************
* FILENAME :        t2fls.h 
*
* DESCRIPTION :
*        GPC algorithm implementation with Type 2 Fuzzy Logic
*
*
* licensed under the Creative Commons Attribution-NonCommercial 4.0 
* 
* AUTHOR :    Romulo Antao        START DATE :    27 MAR 17
*
************************************************************************/

/* Standard includes. */
#include <stdio.h>
#include <limits.h>

/* Hardware includes. */
#include <math.h> 

/* Misc definitions */
#include "t2fls.h"

/* Type 2 Fuzzy Model Variables */
Matrix X;

Matrix M1;
Matrix M2;
Matrix sigma;

Matrix C;
Matrix S;

Matrix fU;
Matrix fL;
Matrix wU;
Matrix wL;
Matrix yU;
Matrix yL;

Matrix PHI;

Matrix aU;
Matrix aL;
Matrix bU;
Matrix bL;
Matrix aU_tilde;
Matrix aL_tilde;

extern Matrix y;
extern Matrix u;

/* Model Parameters */
// float a_tilde[na+1] = {1, -1.8, 0.8};
// float b[nb] = {0.4, 0.6}; 

/* Select just the regressors to be evaluated at the antecedent part */
/* Check better way to accomplish this */
int  aReg[NA] = {0, 2};

int  sU_idx[R], sL_idx[R]; /* vector storing the ascending order sorted indexes */

/* Initialize T2FLS Matrices */
int t2fls_init(void)
{
	matrix_init(&X,NC,1);
	matrix_init(&M1,R,NA);
	matrix_init(&M2,R,NA);
	matrix_init(&sigma,R,NA);

	matrix_init(&C,R,NC);
	matrix_init(&S,R,NC);

	matrix_init(&fU,R,1);
	matrix_init(&fL,R,1);
	matrix_init(&wU,R,1);
	matrix_init(&wL,R,1);
	matrix_init(&yU,R,1);
	matrix_init(&yL,R,1);

	matrix_init(&PHI,NC,1);

	matrix_init(&aU,na+1,1);
	matrix_init(&aL,na+1,1);
	matrix_init(&bU,nb,1);
	matrix_init(&bL,nb,1);

	matrix_init(&aU_tilde,na+2,1);
	matrix_init(&aL_tilde,na+2,1);

	return 0;
}

/* Iterate the Type-2 TS FLS */
float t2fls_iterate(void)
{
	int i, j;
	float uu = 0, ll=0;
	float y_hat; 

	/* Evaluate the membership degree of the inputs on each rule */
	for (i=0;i<R;i++)
	{
		fU.matrix[i][0] = 1;
		fL.matrix[i][0] = 1;

		for (j=0;j<NA;j++)
		{
			gauss_t2mf(PHI.matrix[aReg[j]][0],M1.matrix[i][j],M2.matrix[i][j],sigma.matrix[i][j],&uu,&ll);

			// Min aggregation of the upper and lower firing levels 
			if (uu < fU.matrix[i][0])
			 	fU.matrix[i][0] = uu;

			if (ll < fL.matrix[i][0])
			 	fL.matrix[i][0] = ll;
		}
	}

	/* Get the output of each rule consequent */
	for (i=0;i<R;i++)
	{
		yU.matrix[i][0] = 0;
		yL.matrix[i][0] = 0;
		
		for (j=0;j<NC;j++)
		{
			yU.matrix[i][0] = yU.matrix[i][0] + C.matrix[i][j]*PHI.matrix[j][0] + S.matrix[i][j]*fabs(PHI.matrix[j][0]);
			yL.matrix[i][0] = yL.matrix[i][0] + C.matrix[i][j]*PHI.matrix[j][0] - S.matrix[i][j]*fabs(PHI.matrix[j][0]);
		}
	}

	/* Perform the Type-Reduction KM Algorithm */
	y_hat = typereduction_KM();

	return y_hat;
}

/* Implement Gaussian Type-2 MF with uncertain mean */
void gauss_t2mf(float input, float M1, float M2, float sigma, float *uu, float *ll)
{
	if (input<M1)
	{
		*uu = gaussmf(input,M1,sigma);
		*ll = gaussmf(input,M2,sigma);
	}
	else if (input>M2)
	{
		*uu = gaussmf(input,M2,sigma);
		*ll = gaussmf(input,M1,sigma);
	}
	else if (input>=M1 && input<= (M1+M2)/2)
	{
		*uu = 1;
		*ll = gaussmf(input,M2,sigma);
	}
	else if (input>=(M1+M2)/2 && input<= M2)
	{
		*uu = 1;
		*ll = gaussmf(input,M1,sigma);
	}
	else /* should never happen */
	{
		*uu = -1;
		*ll = -1;
	}
}

/* Implement Gaussian Type-1 MF */
float gaussmf(float x, float c, float sigma)
{
	return expf(-(x-c)*(x-c)/(2*sigma*sigma));
}

/* Implement the Karnik-Mendel Type-Reduction algorithm */
float typereduction_KM(void)
{
	//int   sU_idx[R], sL_idx[R]; /* vector storing the ascending order sorted indexes */
	float l_out=0, r_out=0;      /* bound of the output value */
	float s, s1;				 /* Auxiliary accumulators */
	int   i;

	/* Sort the lower  consequent firing levels in ascending order*/
	sort(&yL, sL_idx);

	/* Initialize accumulators */
	s  = 0;
	s1 = 0;

    /* Calculate an initial estimate for l_out considering L=N-1 (using all lower weights of the antecedents) */
	for (i=0;i<R;i++)
	{
		s = s + yL.matrix[sL_idx[i]][0]*fL.matrix[sL_idx[i]][0];
	}
	for (i=0;i<R;i++)
	{
		s1 = s1 + fL.matrix[i][0];
	}
	l_out = s / s1;

	/*  Iteratively find the L switching point as described in algorithm step 2 (P311 Uncertain rule systems Mendel) */
	for (i=0;i<R;i++)
	{
		s  = s - yL.matrix[sL_idx[i]][0]*fL.matrix[sL_idx[i]][0] + yL.matrix[sL_idx[i]][0]*fU.matrix[sL_idx[i]][0];
		s1 = s1 - fL.matrix[sL_idx[i]][0] + fU.matrix[sL_idx[i]][0];
		if (s/s1<l_out)
		{
			l_out = s/s1;
			wL.matrix[sL_idx[i]][0]=fU.matrix[sL_idx[i]][0]; /* Register the used weight (upper in this case) */
		}
		else
		{
			wL.matrix[sL_idx[i]][0]=fL.matrix[sL_idx[i]][0]; /* Register the used weight (lower in this case) */
		}
	}

	/* Sort the upper  consequent firing levels in ascending order*/
	sort(&yU, sU_idx);

	/* Initialize accumulators */
 	s  = 0;
 	s1 = 0;

    /* Calculate an initial estimate for l_out considering R=0 (using all upper weights of the antecedents) */
	for (i=0;i<R;i++)
	{
		s = s + yU.matrix[sU_idx[i]][0]*fU.matrix[sU_idx[i]][0];
	}
	for (i=0;i<R;i++)
	{
		s1 = s1 + fU.matrix[i][0];
	}
	r_out = s / s1;

	/*  Iteratively find the R switching point as described in algorithm step 2 (P311 Uncertain rule systems Mendel) */
	for (i=0;i<R;i++)
	{
		s  = s - yU.matrix[sU_idx[i]][0]*fU.matrix[sU_idx[i]][0] + yU.matrix[sU_idx[i]][0]*fL.matrix[sU_idx[i]][0];
		s1 = s1 - fU.matrix[sU_idx[i]][0] + fL.matrix[sU_idx[i]][0];
		if (s/s1>r_out)
		{
			r_out = s/s1;
			wU.matrix[sU_idx[i]][0]=fL.matrix[sU_idx[i]][0];  /*Register the used weight (lower in this case)*/ 
		}
		else
		{
			wU.matrix[sU_idx[i]][0]=fU.matrix[sU_idx[i]][0]; /* Register the used weight (upper in this case) */
		}
	}

	return (l_out + r_out)/2;
}

/* Implement the ascending order sort algorithm */
void sort(Matrix *Msrc, int s_list[])
{
	int min = INT_MAX;
	int i, flag[R]={0}, min_idx = 0, s_idx=0;

	/* This algorithm will perform R searches for the minimum */
	for (s_idx=0;s_idx<Msrc->rows;s_idx++)
	{	
		min = INT_MAX;
		for (i=0;i<Msrc->rows;i++)
		{
			if (Msrc->matrix[i][0] < min & flag[i]!=1)
			{
			 	min = Msrc->matrix[i][0];
			  	min_idx = i;
			 	s_list[s_idx] = i;
			}
		}
		flag[min_idx] = 1;
	}
}

/* Configure M1 matrix */
int t2fls_M1(float new_data[][NA])
{
	int i, j;

	for (i=0;i<M1.rows;i++)
	{
		for (j=0;j<M1.cols;j++)
			M1.matrix[i][j] = new_data[i][j];
	}
	return 0;
}

/* Configure M2 matrix */
int t2fls_M2(float new_data[][NA])
{
	int i, j;

	for (i=0;i<M2.rows;i++)
	{
		for (j=0;j<M2.cols;j++)
			M2.matrix[i][j] = new_data[i][j];	
	}
	return 0;

}

/* Configure sigma matrix */
int t2fls_sigma(float new_data[][NA])
{
	int i, j;

	for (i=0;i<sigma.rows;i++)
	{
		for (j=0;j<sigma.cols;j++)
			sigma.matrix[i][j] = new_data[i][j];	
	}
	return 0;
}

/* Configure C matrix */
int t2fls_C(float new_data[][NC])
{
	int i, j;

	for (i=0;i<C.rows;i++)
	{
		for (j=0;j<C.cols;j++)
			C.matrix[i][j] = new_data[i][j];			
	}
	return 0;
}

/* Configure S matrix */
int t2fls_S(float new_data[][NC])
{
	int i, j;

	for (i=0;i<S.rows;i++)
	{
		for (j=0;j<S.cols;j++)
			S.matrix[i][j] = new_data[i][j];			
	}
	return 0;
}

/* shifts new data to the regression vector */
int t2fls_PHI(void)
{
	int i;

	for (i=0;i<nb;i++)
	{
		PHI.matrix[i][0] = u.matrix[i][0];
	}

	for (i=0;i<na;i++)
	{
	 	PHI.matrix[i + nb][0] = y.matrix[i+1][0];
	}

	return 0;
}

int t2fls_linearize(void)
{
	int i, j;
	float sum_wU = 0, sum_wL = 0;

	// Zero the model coefficients from previous iteration
	for (j=0;j<=na;j++)
	{
		aU.matrix[j][0] = 0;
		aL.matrix[j][0] = 0;
	}

	for (j=0;j<nb;j++)
	{
		bU.matrix[j][0] = 0;
		bL.matrix[j][0] = 0;
	}
	
	// Get the normalization weight
	for (i=0;i<R;i++)
	{
		sum_wU = sum_wU + wU.matrix[i][0];
		sum_wL = sum_wL + wL.matrix[i][0];
	}

	// Calculate the coefficients
	aU.matrix[0][0] = 1;
	aL.matrix[0][0] = 1;

	for (i=0;i<R;i++)
	{
		for (j=0;j<nb;j++)
		{
			bU.matrix[j][0] = bU.matrix[j][0] + wU.matrix[i][0]*(C.matrix[i][j] + S.matrix[i][j])/sum_wU;
			bL.matrix[j][0] = bL.matrix[j][0] + wL.matrix[i][0]*(C.matrix[i][j] - S.matrix[i][j])/sum_wL;
		}

		for (j=1;j<=na;j++)
		{
			aU.matrix[j][0] = aU.matrix[j][0] - wU.matrix[i][0]*(C.matrix[i][nb+j-1] + S.matrix[i][nb+j-1])/sum_wU;
			aL.matrix[j][0] = aL.matrix[j][0] - wL.matrix[i][0]*(C.matrix[i][nb+j-1] - S.matrix[i][nb+j-1])/sum_wL;
		}
	}

	aU_tilde.matrix[0][0] = aU.matrix[0][0];
	aL_tilde.matrix[0][0] = aL.matrix[0][0];

	for (i=1;i<na+1;i++)
	{
		aU_tilde.matrix[i][0] = aU.matrix[i][0] - aU.matrix[i-1][0];
		aL_tilde.matrix[i][0] = aL.matrix[i][0] - aL.matrix[i-1][0];
	}

	aU_tilde.matrix[na+1][0] = -aU.matrix[na][0];
	aL_tilde.matrix[na+1][0] = -aL.matrix[na][0];

	return 0;

}

/* Iterate the Type-1 TS FLS */
float t1fls_iterate(void)
{
	int i, j;
	float uu = 0, ll=0;
	float y_hat = 0; 
	float sum_f = 0;

	/* Evaluate the membership degree of the inputs on each rule */
	for (i=0;i<R;i++)
	{
		fU.matrix[i][0] = 1;

		for (j=0;j<NA;j++)
		{
			uu = gaussmf(PHI.matrix[aReg[j]][0], M1.matrix[i][j], sigma.matrix[i][j]);

			// Min aggregation of the firing levels 
			if (uu < fU.matrix[i][0])
			 	fU.matrix[i][0] = uu;
		}
	}

	/* Get the output of each rule consequent */
	for (i=0;i<R;i++)
	{
		yU.matrix[i][0] = 0;
		
		for (j=0;j<NC;j++)
		{
			yU.matrix[i][0] = yU.matrix[i][0] + C.matrix[i][j]*PHI.matrix[j][0];
		}
	}

	/* Merge the outputs of every model */
	for (i=0;i<R;i++)
	{
		y_hat = y_hat + fU.matrix[i][0]*yU.matrix[i][0];
	}
	
	for (i=0;i<R;i++)
	{
		sum_f = sum_f + fU.matrix[i][0];
	}

	return y_hat/sum_f;
}

int t1fls_linearize(void)
{
	int i, j;
	float sum_fU = 0;

	// Zero the model coefficients from previous iteration
	for (j=0;j<=na;j++)
	{
		aU.matrix[j][0] = 0;
	}

	for (j=0;j<nb;j++)
	{
		bU.matrix[j][0] = 0;
	}
	
	// Get the normalization weight
	for (i=0;i<R;i++)
	{
		sum_fU = sum_fU + fU.matrix[i][0];
	}

	// Calculate the coefficients
	aU.matrix[0][0] = 1;

	for (i=0;i<R;i++)
	{
		for (j=0;j<nb;j++)
		{
			bU.matrix[j][0] = bU.matrix[j][0] + fU.matrix[i][0]*C.matrix[i][j]/sum_fU;
		}

		for (j=1;j<=na;j++)
		{
			aU.matrix[j][0] = aU.matrix[j][0] - fU.matrix[i][0]*C.matrix[i][nb+j-1]/sum_fU;
		}
	}

	aU_tilde.matrix[0][0] = aU.matrix[0][0];

	for (i=1;i<na+1;i++)
	{
		aU_tilde.matrix[i][0] = aU.matrix[i][0] - aU.matrix[i-1][0];
	}

	aU_tilde.matrix[na+1][0] = -aU.matrix[na][0];

	return 0;

}