% Generalized Predictive Controller
% Algorithm based on LINEAR ARX MODEL
%
% Last Review
% 18 March 2017
%
% licensed under the Creative Commons Attribution-NonCommercial 4.0 
%
% Romulo Antao
% romuloantao@ua.pt

close all
clear all
clc 

%% Add required folders to the path

cd %current_folder%;
clc

% Add dependencies to path 
toolbox = './deps';
model   = './model';

path(path,toolbox);
path(path,model);

%% Define model
define_reactor

%% Generate reference signal
NEW_REF = 0;

if NEW_REF
    
    % Reference Signal
    setpoint = [29.57 31 27.5 28.5 25 29 29 26 31 31 27.5 28.5 26.5 26.5 32 32 28 28 30 30 26.5 26.5 26.5 26.5 32 32 28 28 30 30 ];    
    
    [r,N] = gen_refSignal(setpoint,150,h);
    
    % Disturbance Signals
    SNR     = -35;
    D_Low   = 0;
    D_High  = 20;
    SMag    = 200;
    n_steps = 8;

    [In_HF_noise, Out_HF_noise, In_LF_disturbance] = gen_noise(SNR,SMag,D_Low,D_High,n_steps,N);

    t = (0:N)'*h;

    save ./data/REFERENCE r N In_HF_noise Out_HF_noise

else
    load ./data/REFERENCE

    t = (0:N)'*h;

end
%% System Parameters

na = 2;
nb = 2;

%% Controller parameters

% GPC
Np = 10;                % Prediction Windows
Nu = 3;                 % Control Window
lambda_u   = 0.1;       % Weight factor for Delta_u

load ./data/FORWARD_ARX.mat
    
%%  Variable initializations

y       = 29.417*ones(N,1);
y_noise = 29.417*ones(N,1);
y_hat   = 29.417*ones(N,1);
u       = 32.5*ones(N,1);

%% Simulate the controller

simEndPoint = N-1-Np;

fprintf(1,'Simulating the Generalized Predictive Controller:      ');

for k = 5 : simEndPoint
    
    % Progress status
    fprintf(1,'\b\b\b\b\b\b  %.2d %%',floor(k/simEndPoint*100));
    
    ut = [t(k-1), u(k-1), In_HF_noise(k-1); ...
          t(k)  , u(k-1), In_HF_noise(k-1)];  
      
    simoptions.InitialState=x0;
    [time,x0,output] = sim(sim_model,[t(k-1) t(k)],simoptions,ut);
    x0 = x0(length(x0),:)';
    
    y(k) = output(length(output),6)';

    % Control error
    ctrl_error(k) = r(k) - y(k); 
    
    % Introduce a disturbance (Comment if undersired)
    if t(k)<1100
        T_in  = 25;
    else
        T_in  = 27;
    end
    
    % Model estimation
    y_hat(k) = th(1)*u(k-1) + th(2)*u(k-2) + th(3)*y(k-1) + th(4)*y(k-2);
   
    % Get next control action   
    u(k) = gpc(y(k:-1:k-na),u(k-1:-1:k-nb),r(k+1:k+Np),[1 -th(3) -th(4)],[th(1) th(2)],Np,Nu,lambda_u);
            
end

fprintf(1,'\b\b\b\b\b\b 100 %% \r');

MSE = sum(ctrl_error(floor(1*simEndPoint/4):simEndPoint).^2)/(length(ctrl_error(floor(1*simEndPoint/4):end)));

du = (u(2:end)-u(1:end-1)).^2;
CTRL_effort = sum(du(floor(1*simEndPoint/4):simEndPoint))/length(du(floor(1*simEndPoint/4):simEndPoint));

fprintf(1,'GPC MSE = %.2d \n\r', MSE);
fprintf(1,'Control Effort = %.2d \n\r', CTRL_effort);


%% Plots
              
figure(1)
x_axis_handler(1) = subplot(3,1,1);
plot(t(1:simEndPoint),[y(1:simEndPoint) r(1:simEndPoint)]), title('Output vs Reference'), grid, ylabel('y'), xlabel('Time (h)');
x_axis_handler(2) = subplot(3,1,2);
plot(t(1:simEndPoint),u(1:simEndPoint),'r'), title('Control Signal'), grid, xlabel('Time (h)');
x_axis_handler(3) = subplot(3,1,3);
plot(t(1:simEndPoint),ctrl_error(1:simEndPoint),'r'), title('Error Signal'), grid, xlabel('Time (h)');

set(gcf,'Name','1 - GPC Results','NumberTitle','off');


linkaxes(x_axis_handler, 'x');      % Link all axes in x

figure(2)
plot(t(1:simEndPoint),[y_hat(1:simEndPoint) y(1:simEndPoint)]),title('Model estimation v System output'),grid,
set(gcf,'Name','2 - Model estimation vs System Output','NumberTitle','off');
legend('y_{hat}','y')

