% Generalized Predictive Controller
% Algorithm based on LINEAR ARX MODEL
%
% licensed under the Creative Commons Attribution-NonCommercial 4.0 
%
% Last Review
% 18 March 2017

close all
clear all
clc 

%% Add required folders to the path

cd %current_folder%;
clc

% Add dependencies to path 
toolbox = './deps';
model   = './model';

path(path,toolbox);
path(path,model);

%% Define model
define_coupled_tank

%% Generate reference signal
NEW_REF = 0;

if NEW_REF
    
    % Reference Signal
%     setpoint = [29.57 31 24.5 24.5 25 29 29 26 31 31 24.5 24.5 26.5 26.5 32 32 28 28 30 30 26.5 26.5 26.5 26.5 32 32 28 28 30 30 ];    
      setpoint = [17.17 25 15 25 30 23 27 25 2 25 25 25 25 25 25 25 ];

    [r,N] = gen_refSignal(setpoint,600,h);
    
    % Disturbance Signals
    SNR     = -35;
    D_Low   = 0;
    D_High  = 20;
    SMag    = 200;
    n_steps = 8;

    [In_HF_noise, Out_HF_noise, In_LF_disturbance] = gen_noise(SNR,SMag,D_Low,D_High,n_steps,N);

    t = (0:N)'*h;

    save ./data/REFERENCE r N In_HF_noise

else
    load ./data/REFERENCE

    t = (0:N)'*h;

end
%% System Parameters

na = 2;
nb = 2;

%% Controller parameters

% GPC
Np = 10;                % Prediction Windows
Nu = 3;                 % Control Window
lambda_u   = 0.2;         % Weight factor for Delta_u
lambda_e   = 0.00;      % SSE weight factor

%RLS
npar = 4;               % Number of estimated parameters
ForgFactor = 0.99;      % Forgetting Factor
P0 = 0.100;               % PMatrix initial value
P = P0 * eye(npar);		% PMatrix initialization
th = 0.1*rand(1,npar);  % Estimated parameters initial values  
Phi = zeros(1, npar);   % Regression vector initial values

load ./data/FORWARD_ARX.mat

%% Set Disturbances
In_HF_noise = In_HF_noise;
In_LF_disturbance = 0*In_LF_disturbance;
Out_LF_disturbance = Out_LF_disturbance;
Out_HF_noise = Out_HF_noise;

%%  Variable initializations

y       = 24.7382*ones(N,1);
y_noise = 24.7382*ones(N,1);
y_hat   = 24.7382*ones(N,1);
u       = 60*ones(N,1);

%% Simulate the controller

simEndPoint = N-1-Np;

fprintf(1,'Simulating the Generalized Predictive Controller:      ');

for k = 5 : simEndPoint
    
    % Progress status
    fprintf(1,'\b\b\b\b\b\b  %.2d %%',floor(k/simEndPoint*100));

    ut = [t(k-1), u(k-1), Out_HF_noise(k-1), In_HF_noise(k-1), Out_LF_disturbance(k-1), In_LF_disturbance(k-1); ...
          t(k)  , u(k-1), Out_HF_noise(k-1), In_HF_noise(k-1), Out_LF_disturbance(k-1), In_LF_disturbance(k-1)];
      
    simoptions.InitialState=x0;
    [time,x0,output] = sim(sim_model,[t(k-1) t(k)],simoptions,ut);
    x0 = x0(length(x0),:)';
    
    y(k) = output(length(output),2)';
    
    if t(k)<5000
%         Out_LF_disturbance(k) = 0*Out_LF_disturbance(k);
        Out_LF_disturbance(k) = 0;
    else
        Out_LF_disturbance(k) =  Out_LF_disturbance(k);
    end
    
    % Control error
    error(k) = r(k) - y(k);
            
    % Model estimation
    y_hat(k) = th(1)*u(k-1) + th(2)*u(k-2) + th(3)*y(k-1) + th(4)*y(k-2);
   
    % Get next control action   
    u(k) = gpc(y(k:-1:k-na),u(k-1:-1:k-nb),r(k+1:k+Np),[1 -th(3) -th(4)],[th(1) th(2)],Np,Nu,lambda_u);
    
    if u(k) < ULow
        u(k) = ULow;
    elseif u(k) > UHigh
        u(k) = UHigh;
    end
        
    % Regression vector update       
    Phi(1)=u(k);
    Phi(2)=u(k-1);
    Phi(3)=y(k);
    Phi(4)=y(k-1);
    
    % For diagnosis purposes  
    for i = 1 : npar
        Theta(k, i) = th(i);
        LongP(k, i) = P(i,i);
    end
        
end

fprintf(1,'\b\b\b\b\b\b 100 %% \r');

MSE = sum(error(floor(1*simEndPoint/4):simEndPoint).^2)/(length(error(floor(1*simEndPoint/4):end)));

du = (u(2:end)-u(1:end-1)).^2;
CTRL_effort = sum(du(floor(1*simEndPoint/4):simEndPoint))/length(du(floor(1*simEndPoint/4):simEndPoint));

fprintf(1,'GPC MSE = %.2d \n\r', MSE);
fprintf(1,'Control Effort = %.2d \n\r', CTRL_effort);


%% Plots
              
figure(1)
x_axis_handler(1) = subplot(3,1,1);
plot(t(1:simEndPoint),[y(1:simEndPoint) r(1:simEndPoint)]), title('Output vs Reference'), grid, ylabel('y'), xlabel('Time (h)');
x_axis_handler(2) = subplot(3,1,2);
plot(t(1:simEndPoint),u(1:simEndPoint),'r'), title('Control Signal'), grid, xlabel('Time (h)');
x_axis_handler(3) = subplot(3,1,3);
plot(t(1:simEndPoint),error(1:simEndPoint),'r'), title('Error Signal'), grid, xlabel('Time (h)');

set(gcf,'Name','1 - GPC Results','NumberTitle','off');



linkaxes(x_axis_handler, 'x');      % Link all axes in x

figure(2)
plot(t(1:simEndPoint),[y_hat(1:simEndPoint) y(1:simEndPoint)]),title('Model estimation v System output'),grid,
set(gcf,'Name','2 - Model estimation vs System Output','NumberTitle','off');
legend('y_{hat}','y')

figure(3)
plot(t(1:simEndPoint),LongP(1:simEndPoint,:)),title('Covariance matrix'), grid,
set(gcf,'Name','3 - System ID - Covariance matrix','NumberTitle','off');

figure(4)
plot(t(1:simEndPoint),Theta(1:simEndPoint,:)),title('Model Parameters'), grid,
set(gcf,'Name','4 - System ID - Model Parameters','NumberTitle','off');
legend('num(1)','num(2)','den(1)','den(2)')

save ./results/ResultsARX

%matlab2tikz('./results/Results1/output_comparison1.tex')
