% licensed under the Creative Commons Attribution-NonCommercial 4.0 
%
% Romulo Antao romuloantao@ua.pt

classdef t2fls < handle
    
  properties
       
      %% System structure configuration
      ant_type;
      cons_type;
      
      %% T2FLS main parameters
      M1;
      M2;
      M;
      sigma;
      sigma1;
      sigma2;
      C_r;
      C_l;
      C;
      S;
      
      REG;     % Regressors considered at antecendent
      R;       % Number of rules
      ant;     % Number of antecedents
       
      %% T2FLS intermediate signals
      ra_ul; %rule activation upper level
      ra_ll; %rule activation lower level
      r_out;
      l_out;
      I2l;
      I2r;
      wr;
      I1r;
      I1l;
      wl;
      y_r;
      y_l;
      R_i;
      L_i;
        
      %% T1FLS intermediate signals (remaining is identical to the Type 2 implementation)
      
      fL;                          % Firing level of each rule
      y_i;                         % Individual rule output
      
      %% RLS parameters for network consequent training

      npar_rls;                    % Number of estimated parameters for each rule

      P0_rls;                      % PMatrix initial value        
      P_rls;                   	   % PMatrix initialization
      th_rls;                      % Estimated parameters 
      Phi_rls;                     % Regression vector
      
      error_rls;                   % RLS estimation error
             
      %% Parameters for data clustering 
      % Constants as suggested by the paper "Simplified Interval Type2 Fuzzy Neural Networks"

      sigma_fixed;
      beta;
    
      % Rule activation minimum threshold to generate new rule
      f_th;

      %GPC variables
       
      Np
      Nu;
      e;
      f;
      g;
       
      lambda_gpc;
      w;
      G;
      F;
      L;
      K;
       
      al_w;
      bl_w;
       
      ar;
      al;
      br;
      bl;
       
      ar_w;
      br_w;
       
      w_params;
      y_params;
      u_params;
            
   end
   
   
   methods
       
       function obj = t2fls(ant_type,cons_type)
          
           if strcmp(ant_type,'type1_ant')
                obj.ant_type = 1;
           else
                obj.ant_type = 2;
           end
           
           if strcmp(cons_type,'type0_cons')
                obj.cons_type = 0;
           else
                obj.cons_type = 1;
           end
            
       end
       
       function obj = rls_init(obj,P0_init)
               
           % RLS parameters for network consequent training

           obj.P0_rls   = P0_init;                 % PMatrix initial value
           
           if obj.cons_type == 1
               
               obj.npar_rls = 2*obj.ant*obj.R;    % Number of parameters to be estimated in each rule
                
               obj.P_rls    = obj.P0_rls * eye(obj.npar_rls);      % PMatrix initialization

               obj.th_rls   = CS2th(obj.C,obj.S,obj.cons_type);    % Estimated parameters initial values

               obj.Phi_rls  = zeros(1,obj.R*obj.npar_rls);    
               
           else
               obj.npar_rls = obj.ant;      % Number of parameters to be estimated

               for i=1:obj.R
                    obj.P_rls(:,:,i) = obj.P0_rls * eye(obj.npar_rls);	% PMatrix initialization
               end

               obj.th_rls   = CS2th(obj.C,obj.S,obj.cons_type);         % Estimated parameters initial values

               obj.Phi_rls  = zeros(1,obj.npar_rls);                    % Regression vector initial values
           
           end
           
       end
       
       function obj = rls(obj,input,output)
           
            % Generate Phi vector
            obj.Phi_rls = input;
            
            % Train the consequents using Recursive Least Squares
            if obj.ant_type == 1
  
                weight = obj.fL/sum(obj.fL);
                for i=1:obj.R
                       [obj.th_rls(i,:), obj.P_rls(:,:,i), obj.error_rls] = rlsdf(obj.th_rls(i,:), obj.Phi_rls, obj.P_rls(:,:,i), output, obj.npar_rls, 0.995, weight(i));                 
                end   
                
            else
                
                if obj.cons_type == 0
                    
                    weight = [0.5*obj.wr/sum(obj.wr)+0.5*obj.wl/sum(obj.wl)];

                    for i=1:obj.R
                           [obj.th_rls(i,:), obj.P_rls(:,:,i), obj.error_rls] = rlsdf(obj.th_rls(i,:), obj.Phi_rls, obj.P_rls(:,:,i), output, obj.npar_rls, 0.995, weight(i));                                         
                    end
                    
                else
                    
                    obj.Phi_rls = phi_matrix(input,obj.wr,obj.wl,obj.R);
                    [obj.th_rls, obj.P_rls, obj.error_rls] = rlsdf_t2(obj.th_rls, obj.Phi_rls, obj.P_rls, output, obj.npar_rls, 0.995);                                         
                
                end
                                             
            end
            
       end

       function obj = backprop(obj,input,y_hat,y,eta)
          
           % perform backpropagation algorithm for antecendent training
           if obj.ant_type == 1
               [obj.M,obj.sigma]=t1tsk_bp(input,y_hat,y,obj.M,obj.sigma,obj.fL,obj.y_i,eta);
           else
               [obj.M,obj.sigma1,obj.sigma2]=t2tsk_bp(input,y_hat,y,obj.M,obj.sigma1,obj.sigma2,...
                                                           obj.r_out,obj.l_out,obj.I2l,obj.I2r,obj.wr,...
                                                           obj.I1l,obj.I1r,obj.wl,obj.y_r,obj.y_l,eta);
           end
       end
       
       function obj = update_network(obj)

           % Update C and S matrices for the next iteration
           [obj.C,obj.S] = th2CS(obj.th_rls,obj.R,obj.ant,obj.cons_type);
               
       end
              
       function [y_hat,yr,yr_i,yl,yl_i] = iterate(obj,input)
           
           if obj.ant_type == 1
               [y_hat,obj.fL,obj.y_i]=tsk_type1(input,obj.M,obj.sigma,obj.C,obj.REG);
               
               yr = y_hat;
               yl = y_hat;
               yr_i = obj.y_i';
               yl_i = obj.y_i';
               
               obj.ra_ul = obj.fL;
               obj.ra_ll = obj.fL;
               
               obj.wl = obj.fL;
               obj.wr = obj.fL;
               
               obj.L_i = 0;
               obj.R_i = 0;
                              
           else
                
               [y_hat,obj.r_out,obj.l_out,obj.ra_ul,obj.ra_ll,obj.I2l,...
                   obj.I2r,obj.wr,obj.I1r,obj.I1l,...
                   obj.wl,obj.y_r,obj.y_l,obj.R_i,obj.L_i]=tsk_type2_uM(input,obj.M1,obj.M2,obj.sigma,obj.C,obj.S,obj.REG);                               
               
               yr = obj.r_out;
               yl = obj.l_out;

               yr_i = obj.y_r';
               yl_i = obj.y_l';
           end
                        
       end
              
       function [u,u_r,u_l] = gpc(obj,u,y,r,du,na,nb,Np,Nu,lambda)
           
           if obj.ant_type == 2
               [u,u_r,u_l] = gpc_type2(u,y,r,du,na,nb,obj.C_r,obj.C_l,obj.wr,obj.wl,Np,Nu,lambda);  
           else
               u = gpc_type1(u,y,r,du,obj.C,obj.fL,na,nb,Np,Nu,lambda);
              
               u_r = u;
               u_l = u;
           end
           
       end

       function [y_pred] = nstep_pred(obj,u,y,du,na,nb,Np)

            if obj.ant_type == 2
                 y_pred = nstep_pred_type2(u,y,du,na,nb,obj.C_r,obj.C_l,obj.wr,obj.wl,Np);
            else
                 y_pred = nstep_pred_type1(u,y,du,obj.C,obj.fL,na,nb,Np);
            end

        end
    
   end
   
end
