% Type 2 Fuzzy Logic Toolbox example
% Control using GPC
%
% 17 March 2017
%
% licensed under the Creative Commons Attribution-NonCommercial 4.0 
%
% Romulo Antao
% romuloantao@ua.pt

close all
clear all
clc 

%% Add required folders to the path

% Add dependencies to path 
toolbox = '../T2FLS_toolbox';
model   = './model';

path(path,toolbox);
path(path,model);

%% Setup the simulink model
define_reactor

NEW_REF    = 0;

ant_type  = 'type2_ant';
cons_type = 'type1_cons'; 

%% Initialize the system structure

% Initialize forward model parameters
plant = t2fls(ant_type,cons_type);

% Load previously trained model structure

if strcmp(ant_type,'type1_ant')
   load ./data/FORWARD_A1C0_1h.mat
else
   load ./data/FORWARD_A1C0_1h.mat
end

if strcmp(ant_type,'type1_ant')
    % TS Type A1C1
    plant.M  = M;
    plant.M1 = M1;
    plant.M2 = M2;
    plant.sigma = sigma;
    plant.C = C;
    plant.S = S;
    plant.C_r = Cr;
    plant.C_l = Cl;
    plant.REG = REG;
    plant.R = R;
    plant.ant = ant;
else
    % TS Type A2C1
    plant.M  = M;
    plant.M1 = M - 0.05*M; % Adjust this parameter
    plant.M2 = M + 0.05*M; % Adjust this parameter
    plant.sigma  = sigma;
    plant.C = C;
    plant.S = S;
    plant.S = 0.12*C; % Adjust this parameter
    plant.C_r = C + S;
    plant.C_l = C - S;
    plant.REG = REG;
    plant.R = R;
    plant.ant = ant;
end
    
% RLS algorithm parameters' initialization
P0_init = 0.100;
plant.rls_init(P0_init);

% Initialize data variables

if NEW_REF
    
    % Reference Signal
    setpoint = [29.57 31 27.5 28.5 25 29 29 26 31 31 27.5 28.5 26.5 26.5 32 32 28 28 30 30 26.5 26.5 26.5 26.5 32 32 28 28 30 30 ];    
    [r,N] = gen_refSignal(setpoint,150,h);

    % Disturbance Signals
    SNR     = -35;
    D_Low   = 0;
    D_High  = 20;
    SMag    = 200;
    n_steps = 8;
    
    [In_HF_noise, Out_HF_noise, In_LF_disturbance] = gen_noise(SNR,SMag,D_Low,D_High,n_steps,N);
    
    out_noise = 0.025*randn(1,N);
    in_noise  = 0.1*randn(1,N);
    
    % Number of plant inputs
    inputs = size(plant.M,2);      
    
    % Initialize data structures
    data    = t2fls_data(inputs, N);
    
    data.r  = r;    
    data.y  = 29.57*ones(N,1);
    data.y_hat = 29.57*ones(N,1); 
    data.u  = 32.5*ones(N,1); 
    data.t  = (0:data.N)'*h;
    data.y_noise = data.y;

    % Generate the Disturbances        
    save ./data/REFERENCE r N In_HF_noise out_noise in_noise

else
    
    load ./data/REFERENCE

    % Number of plant inputs
    inputs = size(plant.M,2);

    % Initialize data structures
    data = t2fls_data(inputs, N);
    
    data.r = r;
    data.y  = 29.57*ones(N,1);
    data.y_hat = 29.57*ones(N,1);
    data.u  = 32.5*ones(N,1); 
    data.t = (0:data.N)'*h;
    data.y_noise = data.y;
end

%% Simulate the controller

Np = 10;
Nu = 3;
lambda_u = 0.01;
lambda_e = 0.000;
simEndPoint = data.N-1-Np;

%%  Variable initializations

fprintf(1,'Simulating the GPC controller: ...      ');
for k = 20 : simEndPoint
    
    %progress status
    fprintf(1,'\b\b\b\b\b\b  %.2d %%',floor(k/simEndPoint*100));
    
    %	plant Output Simulation
    %	ut = [T, U];
    %	READ OUTPUT FROM THE PHYSICAL SYSTEM
        
    ut = [data.t(k-1), data.u(k-1), 0; ...
          data.t(k)  , data.u(k-1), 0];
      
    simoptions.InitialState=x0;
    [time,x0,output] = sim(sim_model,[data.t(k-1) data.t(k)],simoptions,ut);
    x0 = x0(length(x0),:)';
    
    data.y(k) = output(length(output),6);
        
    data.y_noise(k) = data.y(k) + 0.5*out_noise(k);

    % Simulate model output
    
    data.input(k,:) = [data.u(k-1) data.u(k-2) data.y_noise(k-1) data.y_noise(k-2)];
    
    [data.y_hat(k), data.yr(k), data.yr_i(k,:), data.yl(k), data.yl_i(k,:)] = plant.iterate(data.input(k,:));
    
    data.model_error(k) = data.y(k) - data.y_hat(k);
    
    % get the control loop error
    data.ctrl_error(k) = data.r(k) - data.y_noise(k);
    
    % Introduce a disturbance (Comment if undersired)
    if data.t(k)<1100
        T_in  = 25;
    else
        T_in  = 27;
    end
    
    % Get the next control action
    
    na = 2;
    nb = 2;
    
    [data.u(k),data.ur(k),data.ul(k)] = plant.gpc(data.u(k-1:-1:k-nb),...
        data.y_noise(k:-1:k-na),...
        data.r(k+1:k+Np),data.u(k-1)-data.u(k-2),...
        na,nb,Np,Nu,lambda_u);
    
    data.u(k) = data.u(k);
    
    if data.u(k) < ULow
        data.u(k) = ULow;
    elseif data.u(k) > UHigh
        data.u(k) = UHigh;
    end
              
end

fprintf(1,'\b\b\b\b\b\b100 %% \r');
MSE = sum(data.ctrl_error(floor(1*simEndPoint/4):simEndPoint).^2)/(length(data.ctrl_error(floor(1*simEndPoint/4):simEndPoint)));

du = (data.u(2:end)-data.u(1:end-1)).^2;
CTRL_effort = sum(du(floor(1*simEndPoint/4):simEndPoint))/length(du(floor(1*simEndPoint/4):simEndPoint));

MSE_model = sum(data.model_error(floor(1*simEndPoint/4):simEndPoint).^2)/length(data.model_error(floor(1*simEndPoint/4):simEndPoint));

fprintf(1,'GPC MSE = %.2d \n\r', MSE);
fprintf(1,'Control Effort = %.2d \n\r', CTRL_effort);
fprintf(1,'Model MSE = %.2d \n\r', MSE_model);

%% Plots

figure(1)
x_axis_handler_1(1) = subplot(2,1,1);
plot(data.t(1:simEndPoint), [data.y_hat(1:simEndPoint) data.y(1:simEndPoint)]), title('Model vs Plant output'), grid, ylabel('Output'), xlabel('Time (h)');
hold on 
legend('Output-hat','Output')
x_axis_handler_1(2) = subplot(2,1,2);
plot(data.t(1:simEndPoint), data.model_error(1:simEndPoint),'r'), title('Error'), grid, xlabel('Time (h)');
set(gcf,'Name','1 - GPC - Model vs Plant Output','NumberTitle','off');

linkaxes(x_axis_handler_1, 'x');      % Link all axes in x

figure(2)
x_axis_handler_2(1) = subplot(311);
plot(data.t(1:simEndPoint), [data.y(1:simEndPoint) data.r(1:simEndPoint)]), title('System output vs Control reference'), grid, xlabel('Time (h)'), ylabel('Amplitude')
legend('Output','Reference')

x_axis_handler_2(2) = subplot(312);
plot(data.t(1:simEndPoint), data.u(1:simEndPoint)),title('Control Signal'), grid, xlabel('Time (h)'), ylabel('Amplitude')

x_axis_handler_2(3) = subplot(313);
plot(data.t(1:simEndPoint), data.ctrl_error(1:simEndPoint)),title('Control Error'), grid, xlabel('Time (h)'), ylabel('Amplitude')

set(gcf,'Name','2 - GPC - Control results (1)','NumberTitle','off');

linkaxes(x_axis_handler_2, 'x');      % Link all axes in x

% Save results
if strcmp(ant_type,'type1_ant')
    save ./results/ResultsA1C0_1h
else
    save ./results/ResultsA2C1_1h
end