% Open Loop simulation of the Stirred Tank Reactor process
%
% 17 March 2017
%
% licensed under the Creative Commons Attribution-NonCommercial 4.0 
%
% Romulo Antao
% romuloantao@ua.pt

% NOTE:
% Simulation of second order system using a stair test signal
% To obtain a proper training sequence for TSK model extraction,
% a long test is required to ensure the training of the model at different 
% operation points.
%
close all
clear all
clc 

%% Add required folders to the path

% Add dependencies to path 
model = './model';

path(path,model);

%% Setup the simulink model

define_reactor

%% Control signal parameters

nelemref = 15;          % Number of steps to test
u1 = floor(  0*10 + ( 50*10 -   0*10)*rand(3*nelemref,1))/10;
u2 = floor( 50*10 + (100*10 -  50*10)*rand(2*nelemref,1))/10;
u3 = floor(100*10 + (150*10 - 100*10)*rand(nelemref,1))/10;

setpoint = [18;u1;u2;u3];

[u,N] = gen_uSignal(setpoint,200,h);

%%  Variable initializations

y = zeros(N,1);
y_noise = zeros(N,1);
t = zeros(N,1);

%%	Open loop actuation

simEndPoint = N;

fprintf(1,'Open Loop actuation over the system...      ');
tic
for k = 3 : simEndPoint
    
    %progress status
    fprintf(1,'\b\b\b\b\b\b  %.2d %%',floor(k/simEndPoint*100));
    
	t(k) = t(k-1) + h;					% Continuous time update
    
%	Plant Output Simulation
%	ut = [T, U, LOAD];
%	READ OUTPUT FROM THE PHYSICAL SYSTEM

    ut = [t(k-1), u(k-1), 0; ...
          t(k)  , u(k-1), 0];
	simoptions.InitialState=x0;
	[time,x0,output] = sim(sim_model,[t(k-1) t(k)],simoptions,ut);
	x0 = x0(length(x0),:)';
    
	y(k) = output(length(output),6)';
    
end

%generate random noise

y_noise = 0.1*randn(size(y));

fprintf(1,'\b\b\b\b\b\b100 %% \r\n');
toc
%% Plot the results

figure(1)

x_axis_handler(1) = subplot(211);
plot(t(1:simEndPoint),y(1:simEndPoint)), title('Output Signal'), xlabel('Time (h)'), ylabel('Amplitude')
axis([0 t(end) 1.1*min(y) 1.1*max(y)])
x_axis_handler(2) = subplot(212);
plot(t(1:simEndPoint),u(1:simEndPoint)),title('Control Signal') ,xlabel('Time (h)'), ylabel('Amplitude')
axis([0 t(end) 1.1*min(u) 1.1*max(u)])

linkaxes(x_axis_handler, 'x');      % Link all axes in x

%% Save some important data !!!

save ./data/MODEL_SIGNALS_1h N h y u y_noise

% EOF : open_model.m





