% Type 2 Fuzzy Logic Toolbox example
% Forward model synthesis
%
% 17 March 2017
%
% licensed under the Creative Commons Attribution-NonCommercial 4.0 
%
% Romulo Antao
% romuloantao@ua.pt

close all
clear all
clc 

%% Add required folders to the path

% Set current folder
clc

% Add dependencies to path 
toolbox = '../T2FLS_toolbox';
model   = './model';

path(path,toolbox);
path(path,model);

%% Setup the simulink model
define_coupled_tank

%% Training related variables
load ./data/MODEL_SIGNALS_2s.mat

%% Train the model using the open loop data

for k = 3 : N-1        
    input(k,:) = [u(k-1) u(k-2) y(k-1) y(k-2)];
    output(k,:) = y(k);
end

[N, inputs] = size(input);

% Initialize input/output data structure
data = t2fls_data(inputs, N);

data.input = input;
data.y = output;

% Initialize the model
plant = t2fls(ant_type,cons_type);

% Start with random data for the training of the Type-1 Model
% Perform Subtractive Clustering on the training dataset
[M,SIGMA]    = subclust(input,0.4, []);

plant.R      = size(M,1);
plant.ant    = size(M,2);
plant.REG    = [1,3];
plant.M      = M;
plant.M1     = M - 0.05*M;
plant.M2     = M + 0.05*M;
plant.sigma  = repmat(SIGMA,plant.R,1);
plant.C      = 0.1*rand(plant.R,plant.ant);
plant.S      = 0;
plant.C_r    = 0.1*rand(plant.R,plant.ant);
plant.C_l    = 0.1*rand(plant.R,plant.ant);

M       = plant.M;
M1      = plant.M1;
M2      = plant.M2;
sigma   = plant.sigma;
C       = plant.C;
S       = plant.S;
Cr      = plant.C_r;
Cl      = plant.C_l;
REG     = plant.REG;
R       = plant.R;
ant     = plant.ant;


% RLS algorithm parameters' initialization
P0_init = 100;  % Use 0.1 for Type-2 Model - More confidence on the starting point
plant.rls_init(P0_init);

%% Train model
simEndPoint = data.N;
data.t      = (0:data.N)'*h;
    
% Training epoches
tic
    
fprintf(1,'ID the fuzzy model: ...      ');

for k = 4 : simEndPoint
    
    fprintf(1,'\b\b\b\b\b\b  %.2d %%',floor(k/simEndPoint*100));
    
    % Continuous time update
    data.t(k) = data.t(k-1) + h;
    
    % Make one iteration of the current model
    [data.y_hat(k), data.yr(k), data.yr_i(k,:), data.yl(k),data.yl_i(k,:)] = plant.iterate(data.input(k,:));
    
    % Get estimation error
    data.model_error(k) = data.y(k) - data.y_hat(k);
    
    % Train the consequents using Recursive Least Squares
    plant.rls(data.input(k,:),data.y(k));
    
    % Update T2FLS parameters after training
    plant.update_network();
       
end

fprintf(1,'\b\b\b\b\b\b100 %% \n');

MSE = data.model_error(floor(simEndPoint/10):end)'*data.model_error(floor(simEndPoint/10):end)/(2*size(data.model_error(floor(simEndPoint/10):end),1));

fprintf(1,'Forward Model MSE = %.2d \n\r', MSE);

toc       
  
%% Save the model parameters

M  = plant.M;
M1 = plant.M1;
M2 = plant.M2;
sigma  = plant.sigma;
C = plant.C;
S = plant.S;
Cr = plant.C + plant.S;
Cl = plant.C - plant.S;
R = plant.R;
ant = plant.ant;

save ./data/FORWARD_A1C0 M M1 M2 sigma Cl Cr C S R ant REG

%% Plots

figure(1)
x_axis_handler(1) = subplot(2,1,1);
plot(data.t(1:simEndPoint), data.y_hat(1:simEndPoint)), title('Estimated model vs Reference Model'), grid, ylabel('Output'), xlabel('Time (s)');
hold on 
plot(data.t(1:simEndPoint), data.y(1:simEndPoint),'r')
legend('y-hat','y')
x_axis_handler(2) = subplot(2,1,2);
plot(data.t(1:simEndPoint), data.model_error(1:simEndPoint),'r'), title('Error'), grid, xlabel('Time (s)');
set(gcf,'Name','1 - Forward Model Training - 1 ','NumberTitle','off');

linkaxes(x_axis_handler, 'x');      % Link all axes in x
