% Type 2 Fuzzy Logic Toolbox example
% Control using GPC
%
% 17 March 2017
%
% licensed under the Creative Commons Attribution-NonCommercial 4.0 
%
% Romulo Antao
% romuloantao@ua.pt

close all
clear all
clc

%% Add required folders to the path

% Set current folder
clc

% Add dependencies to path 
toolbox = '../T2FLS_toolbox';
model   = './model';

path(path,toolbox);
path(path,model);

%% Setup the simulink model
define_coupled_tank

NEW_REF    = 0;

ant_type  = 'type2_ant';
cons_type = 'type1_cons'; 

%% Initialize the system structure

% Initialize forward model parameters
plant = t2fls(ant_type,cons_type);

% Load previously trained model structure

if strcmp(ant_type,'type1_ant')
     load ./data/FORWARD_A1C0.mat
else
     load ./data/FORWARD_A1C0.mat
end

if strcmp(ant_type,'type1_ant')
    % TS Type A1C0
    plant.M  = M;
    plant.M1 = M1;
    plant.M2 = M2;
    plant.sigma = sigma;
    plant.C = C;
    plant.S = S;
    plant.C_r = Cr;
    plant.C_l = Cl;
    plant.REG = REG;
    plant.R = R;
    plant.ant = ant;
else
    % TS Type A2C1
    plant.M  = M;
    plant.M1 = M - 0.05*M; % Adjust this parameter
    plant.M2 = M + 0.05*M; % Adjust this parameter
    plant.sigma  = sigma;
    plant.C = C;
    plant.S = 0.08*C;  % Adjust this parameter
    plant.C_r = plant.C + plant.S;
    plant.C_l = plant.C - plant.S;
    plant.REG = REG;
    plant.R = R;
    plant.ant = ant;
end
    

% RLS and RLM algorithms parameters' initialization
P0_init = 100;
plant.rls_init(P0_init);

% Initialize data variables

if NEW_REF
    
    setpoint = [17.17 25 15 25 30 23 27 20 20 25 15 25 30 23 27 ];

    [r,N] = gen_refSignal(setpoint,600,h);
        
    % Number of plant inputs
    inputs = size(plant.M,2);

    % Initialize data structures
    data    = t2fls_data(inputs, N);

    data.r  = r;
    data.y  = 24.7382*ones(N,1);
    data.y_hat = 24.7382*ones(N,1);
    data.u  = 60*ones(N,1); 
    data.t = (0:data.N)'*h;
    
    % Generate the Disturbances
    
    % SNR -48db (2bit in a  10bit ADC)
    SNR = -48;
    n_steps = 3;
    D_Low = 0; D_High = 0;
    [In_HF_noise, Out_HF_noise, In_LF_disturbance] = gen_noise(SNR,UHigh,D_Low,D_High,n_steps,N);

    In_HF_noise = 0.5*randn(size(data.r));
    Out_HF_noise = 0.01*randn(size(data.r));
    
    Out_LF_disturbance = (8*sin(2*pi*1/800*data.t)+8);
       
    save ./data/REFERENCE r In_HF_noise Out_HF_noise In_LF_disturbance Out_LF_disturbance N

else
    
    load ./data/REFERENCE

    % Number of plant inputs
    inputs = size(plant.M,2);

    % Initialize data structures
    data.r = r;
    data.y  = 24.738*ones(N,1);
    data.y_hat = 24.738*ones(N,1);
    data.u  = 60*ones(N,1); 
    data.t = (0:N)'*h;
    data.N = N;

end

In_HF_noise = In_HF_noise;
In_LF_disturbance = 0*In_LF_disturbance;
Out_LF_disturbance = Out_LF_disturbance;
Out_HF_noise = Out_HF_noise;

%% Simulate the controller

Np = 9;
Nu = 3;
lambda = 0.5;

simEndPoint = data.N-1-Np;

% Training epoches
tic

fprintf(1,'Simulating the GPC controller: ...      ');
for k = 20 : simEndPoint
    
    %progress status
    fprintf(1,'\b\b\b\b\b\b  %.2d %%',floor(k/simEndPoint*100));
    
    %	plant Output Simulation
    %	ut = [T, U];
    %	READ OUTPUT FROM THE PHYSICAL SYSTEM
    
    ut = [data.t(k-1), data.u(k-1), Out_HF_noise(k-1), In_HF_noise(k-1), Out_LF_disturbance(k-1), In_LF_disturbance(k-1); ...
          data.t(k)  , data.u(k-1), Out_HF_noise(k-1), In_HF_noise(k-1), Out_LF_disturbance(k-1), In_LF_disturbance(k-1)];
    
    simoptions.InitialState=x0;
    [time,x0,output] = sim(sim_model,[data.t(k-1) data.t(k)],simoptions,ut);
    x0 = x0(length(x0),:)';
    
    data.y(k) = output(length(output),2)';
    
    % Simulate model output
    
    data.input(k,:) = [data.u(k-1) data.u(k-2) data.y(k-1) data.y(k-2)];
    
    [data.y_hat(k), data.yr(k), data.yr_i(k,:), data.yl(k), data.yl_i(k,:)] = plant.iterate(data.input(k,:));
    
    data.model_error(k) = data.y(k) - data.y_hat(k);
    
    % get the control loop error
    data.ctrl_error(k) = data.r(k) - data.y(k);
    
    if data.t(k)<5000
        Out_LF_disturbance(k) = 0;
    else
         Out_LF_disturbance(k) = 20;
    end
    
    % Get the next control action
    
    na = 2;
    nb = 2;
    
    [data.u(k),data.ur(k),data.ul(k)] = plant.gpc(data.u(k-1:-1:k-nb),...
        data.y(k:-1:k-na),...
        data.r(k+1:k+Np),data.u(k-1)-data.u(k-2),...
        na,nb,Np,Nu,lambda);
    
    if data.u(k) < ULow
        data.u(k) = ULow;
    end
    if data.u(k) > UHigh
        data.u(k) = UHigh;
    end
        
end

fprintf(1,'\b\b\b\b\b\b100 %% \r');
MSE = sum(data.ctrl_error(floor(1*simEndPoint/4):simEndPoint).^2)/(length(data.ctrl_error(floor(1*simEndPoint/4):simEndPoint)));

du = (data.u(2:end)-data.u(1:end-1)).^2;
CTRL_effort = sum(du(floor(1*simEndPoint/4):simEndPoint))/length(du(floor(1*simEndPoint/4):simEndPoint));

MSE_model = sum(data.model_error(floor(1*simEndPoint/4):simEndPoint).^2)/length(data.model_error(floor(1*simEndPoint/4):simEndPoint));

fprintf(1,'GPC MSE = %.2d \n\r', MSE);
fprintf(1,'Control Effort = %.2d \n\r', CTRL_effort);
fprintf(1,'Model MSE = %.2d \n\r', MSE_model);

toc

%% Plots

figure(1)
x_axis_handler_1(1) = subplot(2,1,1);
plot(data.t(1:simEndPoint), [data.y_hat(1:simEndPoint) data.y(1:simEndPoint)]), title('Model vs Plant output'), grid, ylabel('Output'), xlabel('Time (s)');
hold on 
legend('Output-hat','Output')
x_axis_handler_1(2) = subplot(2,1,2);
plot(data.t(1:simEndPoint), data.model_error(1:simEndPoint),'r'), title('Error'), grid, xlabel('Time (h)');
set(gcf,'Name','1 - GPC - Model vs Plant Output','NumberTitle','off');

linkaxes(x_axis_handler_1, 'x');      % Link all axes in x

figure(2)
x_axis_handler_2(1) = subplot(311);
plot(data.t(1:simEndPoint), [data.y(1:simEndPoint) data.r(1:simEndPoint)]), title('System output vs Control reference'), grid, xlabel('Time (s)'), ylabel('Amplitude')
legend('Output','Reference')

x_axis_handler_2(2) = subplot(312);
plot(data.t(1:simEndPoint), data.u(1:simEndPoint)),title('Control Signal'), grid, xlabel('Time (s)'), ylabel('Amplitude')

x_axis_handler_2(3) = subplot(313);
plot(data.t(1:simEndPoint), data.ctrl_error(1:simEndPoint)),title('Control Error'), grid, xlabel('Time (s)'), ylabel('Amplitude')

set(gcf,'Name','2 - GPC - Control results (1)','NumberTitle','off');
